from django.utils.deconstruct import deconstructible
from django.utils.translation import gettext_lazy as _

from vies.types import VATIN


@deconstructible
class VATINValidator:
    """Validator for European VIES VAT Identification Number."""

    message = _("Not a valid European VAT number.")
    code = "invalid"

    def __init__(self, verify=True, validate=False):
        if not (verify or validate):
            raise ValueError('"verify" and "validate" can not both be false.')
        self.verify = verify
        self.validate = validate

    def __call__(self, value):
        if isinstance(value, str):
            value = VATIN.from_str(value)
        if self.verify:
            value.verify()
        if self.validate:
            value.validate()



"""Validate and store VAT Information Exchange System (VIES) data in Django."""

import logging

from . import _version  # noqa

logger = logging.getLogger("vies")

__version__ = _version.__version__
VERSION = _version.VERSION_TUPLE

VIES_WSDL_URL = "https://ec.europa.eu/taxation_customs/vies/checkVatService.wsdl"  # NoQA
VATIN_MAX_LENGTH = 14



from django import forms

from vies.types import VATIN, VIES_COUNTRY_CHOICES
from vies.validators import VATINValidator

from .. import VATIN_MAX_LENGTH
from .widgets import VATINHiddenWidget, VATINWidget


class VATINField(forms.MultiValueField):
    hidden_widget = VATINHiddenWidget
    widget = VATINWidget

    def __init__(self, choices=VIES_COUNTRY_CHOICES, *args, **kwargs):
        max_length = kwargs.pop("max_length", VATIN_MAX_LENGTH)

        kwargs["widget"] = self.widget(choices=choices)
        kwargs.setdefault("validators", [VATINValidator()])

        fields = (
            forms.ChoiceField(required=False, choices=choices),
            forms.CharField(required=False, max_length=max_length),
        )

        # In Django 1.11+, ignore the `empty_value` parameter added by the
        # `CharField` superclass at the end of `VATINField.formfield`.
        kwargs.pop("empty_value", None)

        super().__init__(fields=fields, *args, **kwargs)

    def compress(self, data_list):
        if data_list:
            return VATIN(*data_list)



import contextlib
import re

from django import forms
from django.forms.widgets import HiddenInput

from vies.types import VIES_COUNTRY_CHOICES

EMPTY_VALUES = (None, "")


class VATINWidget(forms.MultiWidget):
    def __init__(self, choices=VIES_COUNTRY_CHOICES, attrs=None):
        widgets = (forms.Select(choices=choices), forms.TextInput())
        super().__init__(widgets, attrs)

    def value_from_datadict(self, data, files, name):
        value = super().value_from_datadict(data, files, name)
        country, code = value
        if code not in EMPTY_VALUES:
            if country in EMPTY_VALUES:
                with contextlib.suppress(ValueError):
                    # ex. code="FR09443710785", country="".
                    empty, country, code = re.split("([a-zA-Z].)", code)
            else:
                # ex. code ="FR09443710785", country="FR".
                re_code = re.compile(rf"^{country}(\d+)$")
                if re_code.match(code):
                    code = code.replace(country, "", 1)
            with contextlib.suppress(AttributeError):
                country = country.upper()
        return [country, code]

    def format_output(self, rendered_widgets):
        return f"{rendered_widgets[0]}&nbsp;{rendered_widgets[1]}"

    def decompress(self, value):
        if value:
            try:
                country, code = value
            except ValueError:
                country = value[:2]
                code = value[2:]
            return country, code
        return None, None


class VATINHiddenWidget(VATINWidget):
    """Widget that splits vat input into two <input type="hidden"> inputs."""

    def __init__(self, attrs=None):
        widgets = (HiddenInput(attrs=attrs), HiddenInput(attrs=attrs))
        super(VATINWidget, self).__init__(widgets, attrs)



from django.db.models import CharField
from django.utils.translation import gettext_lazy as _

from . import VATIN_MAX_LENGTH, forms


class VATINField(CharField):
    """Database field for European VIES VAT Identification Number.

    This field stores and validates VATINs.

    Example:
        class MyModel(models.Model):
            vat = VATINField(_('EU VAT ID'))

    """

    description = _("A VIES VAT field.")

    def __init__(self, *args, **kwargs):
        kwargs.setdefault("max_length", VATIN_MAX_LENGTH)
        super().__init__(*args, **kwargs)

    def formfield(self, **kwargs):
        kwargs.setdefault("form_class", forms.VATINField)
        return super().formfield(**kwargs)



import re

from django.core.exceptions import ValidationError
from django.utils.functional import cached_property
from django.utils.translation import gettext
from zeep import Client

from vies import VIES_WSDL_URL, logger


def dk_format(v):
    return f"{v[:4]} {v[4:6]} {v[6:8]} {v[8:10]}"


def fr_format(v):
    return f"{v[:4]} {v[4:]}"


VIES_OPTIONS = {
    "AT": ("Austria", re.compile(r"^ATU\d{8}$")),
    "BE": ("Belgium", re.compile(r"^BE(0|1)\d{9}$")),
    "BG": ("Bulgaria", re.compile(r"^BG\d{9,10}$")),
    "HR": ("Croatia", re.compile(r"^HR\d{11}$")),
    "CHE": ("Switzerland", re.compile(r"^CHE\d{9}$")),
    "CY": ("Cyprus", re.compile(r"^CY\d{8}[A-Z]$")),
    "CZ": ("Czech Republic", re.compile(r"^CZ\d{8,10}$")),
    "DE": ("Germany", re.compile(r"^DE\d{9}$")),
    "DK": ("Denmark", re.compile(r"^DK\d{8}$"), dk_format),
    "EE": ("Estonia", re.compile(r"^EE\d{9}$")),
    "EL": ("Greece", re.compile(r"^EL\d{9}$")),
    "ES": ("Spain", re.compile(r"^ES[A-Z0-9]\d{7}[A-Z0-9]$")),
    "FI": ("Finland", re.compile(r"^FI\d{8}$")),
    "FR": ("France", re.compile(r"^FR[A-HJ-NP-Z0-9][A-HJ-NP-Z0-9]\d{9}$"), fr_format),
    "HU": ("Hungary", re.compile(r"^HU\d{8}$")),
    "IE": ("Ireland", re.compile(r"^IE\d[A-Z0-9\+\*]\d{5}[A-Z]{1,2}$")),
    "IT": ("Italy", re.compile(r"^IT\d{11}$")),
    "LT": ("Lithuania", re.compile(r"^LT(\d{9}|\d{12})$")),
    "LU": ("Luxembourg", re.compile(r"^LU\d{8}$")),
    "LV": ("Latvia", re.compile(r"^LV\d{11}$")),
    "MT": ("Malta", re.compile(r"^MT\d{8}$")),
    "NL": ("The Netherlands", re.compile(r"^NL\d{9}B\d{2}$")),
    "PL": ("Poland", re.compile(r"^PL\d{10}$")),
    "PT": ("Portugal", re.compile(r"^PT\d{9}$")),
    "RO": ("Romania", re.compile(r"^RO\d{2,10}$")),
    "SE": ("Sweden", re.compile(r"^SE\d{10}01$")),
    "SI": ("Slovenia", re.compile(r"^SI\d{8}$")),
    "SK": ("Slovakia", re.compile(r"^SK\d{10}$")),
    "XI": ("Northern Ireland", re.compile(r"^XI\d{9}$")),
}

VIES_COUNTRY_CHOICES = sorted(
    (("", "--"),) + tuple((key, key) for key, value in VIES_OPTIONS.items())
)

MEMBER_COUNTRY_CODES = VIES_OPTIONS.keys()


class VATIN:
    """Object wrapper for the european VAT Identification Number."""

    def __init__(self, country_code, number):
        self.country_code = country_code
        self.number = number

    def __str__(self):
        unformated_number = f"{self.country_code}{self.number}"

        country = VIES_OPTIONS.get(self.country_code, {})
        if len(country) == 3:
            return country[2](unformated_number)
        return unformated_number

    def __repr__(self):
        return f"<VATIN {self.__str__()}>"

    def get_country_code(self):
        return self._country_code

    def set_country_code(self, value):
        self._country_code = value.upper()

    country_code = property(get_country_code, set_country_code)

    def get_number(self):
        return self._number

    def set_number(self, value):
        self._number = value.upper().replace(" ", "")

    number = property(get_number, set_number)

    @cached_property
    def data(self):
        """VIES API response data."""
        client = Client(VIES_WSDL_URL)
        try:
            return client.service.checkVat(self.country_code, self.number)
        except Exception as e:
            logger.exception(e)
            raise

    def is_valid(self):
        try:
            self.verify()
            self.validate()
        except ValidationError:
            return False
        else:
            return True

    def verify_country_code(self):
        if not re.match(r"^[a-zA-Z]", self.country_code):
            msg = gettext("%s is not a valid ISO_3166-1 country code.")
            raise ValidationError(msg % self.country_code)
        if self.country_code not in MEMBER_COUNTRY_CODES:
            msg = gettext("%s is not a european member state.")
            raise ValidationError(msg % self.country_code)

    def verify_regex(self):
        country = dict(
            map(
                lambda x, y: (x, y),
                ("country", "validator", "formatter"),
                VIES_OPTIONS[self.country_code],
            )
        )
        if not country["validator"].match(f"{self.country_code}{self.number}"):
            msg = gettext("%s does not match the country's VAT ID specifications.")
            raise ValidationError(msg % self)

    def verify(self):
        self.verify_country_code()
        self.verify_regex()

    def validate(self):
        if not self.data.valid:
            msg = gettext("%s is not a valid VATIN.")
            raise ValidationError(msg % self)

    @classmethod
    def from_str(cls, value):
        """Return a VATIN object by given string."""
        return cls(value[:2].strip(), value[2:].strip())
